# Decoder for IR remote OpenLaserTag (Sony SIRC) protocol .
# Tomas Krejci [Njord]

from utime import ticks_us, ticks_diff
from olt_lib.ir_rx import IR_RX

# Bit reverse a 32 bit value
def rbit32(v):
    v = (v & 0x0000ffff) << 16 | (v & 0xffff0000) >> 16
    v = (v & 0x00ff00ff) << 8 | (v & 0xff00ff00) >> 8
    v = (v & 0x0f0f0f0f) << 4 | (v & 0xf0f0f0f0) >> 4
    v = (v & 0x33333333) << 2 | (v & 0xcccccccc) >> 2
    return (v & 0x55555555) << 1 | (v & 0xaaaaaaaa) >> 1


class SONY_ABC(IR_RX):  # Abstract base class
    def __init__(self, pin, bits, callback, *args):
        # 24 bit block has 50 edges and lasts <= 39ms nominal. Add 4ms to time
        # for tolerances except in 24 bit case where timing is tight with a
        # repeat period of 45ms.
        t = int(3 + bits * 1.8) + (1 if bits >= 20 else 4)
        super().__init__(pin, 2 + bits * 2, t, callback, *args)
        self._addr = 0
        self._bits = 24

    def decode(self, _):
        try:
            nedges = self.edge  # No. of edges detected
            self.verbose and print('nedges', nedges)
            if nedges > 50:
                raise RuntimeError(self.OVERRUN)
            bits = (nedges - 2) // 2
            if nedges not in (26, 32, 42, 50) or bits > self._bits:
                raise RuntimeError(self.BADBLOCK)
            self.verbose and print('SIRC {}bit'.format(bits))
            width = ticks_diff(self._times[1], self._times[0])
            if not 1800 < width < 3000:  # 2.4ms leading mark for all valid data
                raise RuntimeError(self.BADSTART)
            width = ticks_diff(self._times[2], self._times[1])
            if not 350 < width < 1000:  # 600μs space
                raise RuntimeError(self.BADSTART)

            byte1 = 0
            byte2 = 0
            byte3 = 0
            packet = 0
            val = 0  # Data received, LSB 1st
            x = 2
            bit = 1
            while x <= nedges - 2:
                if ticks_diff(self._times[x + 1], self._times[x]) > 900:
                    val |= bit
                bit <<= 1
                x += 2
            val = rbit32(val)
            val >>= 8
            packet = val
            byte3 = val & 0xff  # 8 bit command
            val >>= 8
            byte2 = val & 0xff  # 8 bit command
            val >>= 8
            byte1 = val & 0xff  # 8 bit command
            #if nedges < 42:
            #    addr = val & 0xff  # 5 or 8 bit addr
            #    val = 0
            #else:
            #    addr = val & 0x1f  # 5 bit addr
            #    val >>= 5  # 8 bit extended
        except RuntimeError as e:
            packet = e.args[0]
            byte1 = 0
            byte2 = 0
            byte3 = 0
            #packet = 0
        self.do_callback(byte1, byte2, byte3, packet)

class SONY_12(SONY_ABC):
    def __init__(self, pin, callback, *args):
        super().__init__(pin, 12, callback, *args)

class SONY_15(SONY_ABC):
    def __init__(self, pin, callback, *args):
        super().__init__(pin, 15, callback, *args)

class SONY_20(SONY_ABC):
    def __init__(self, pin, callback, *args):
        super().__init__(pin, 20, callback, *args)

class LT_24(SONY_ABC):
    def __init__(self, pin, callback, *args):
        super().__init__(pin, 24, callback, *args)
